#
#  engine
#
#  Playout Daemon for autoradio project
#
#
#  Copyright (C) 2017-2018 Gottfried Gaisbauer <gottfried.gaisbauer@servus.at>
#
#  This file is part of engine.
#
#  engine is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  any later version.
#
#  engine is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with engine. If not, see <http://www.gnu.org/licenses/>.
#

import simplejson
import decimal
import traceback
import sqlalchemy
import datetime
import logging

from flask import request, render_template, Flask, Response

from libraries.database.database import APP
from libraries.database.broadcasts import TrackService, Schedule, ScheduleEntry

app = Flask(__name__)

def alchemyencoder(obj):
    """JSON encoder function for SQLAlchemy special classes."""
    if isinstance(obj, datetime.date):
        return obj.isoformat()
    elif isinstance(obj, decimal.Decimal):
        return float(obj)
    elif isinstance(obj, sqlalchemy.orm.state.InstanceState):
        return ""
    elif isinstance(obj, Schedule):
        return simplejson.dumps([obj._asdict()], default=alchemyencoder)
    else:
        return str(obj)


#class EndpointAction(object):
#    __name__ = ""

#    def __init__(self, action, content_type, name):
#        self.__name__ = name
#        self.action = action
#        self.response = Response(status=200, headers={"Content-Type": content_type})

#    def __call__(self, *args):
#        self.response.response = self.action()
#        return self.response


#class FlaskWrapper(object):
#    app = None

#    def __init__(self, name):
#        self.app = Flask(name)

#    def run(self):
#        self.app.run()

#    def add_endpoint(self, endpoint=None, endpoint_name=None, handler=None, methods=None, defaults=None, content_type="text/html"):
        # endpoint = rule
        #self.app.add_url_rule()
#        self.app.add_url_rule(endpoint, view_func=EndpointAction(handler, content_type, __name__), methods=methods, defaults=defaults)

class Routes:
    error = None
    scheduler = None
    messenger = None
    flask_wrapper = None
    lqs_communicator = None

    def __init__(self, scheduler, lqs_communicator, messenger):
        self.scheduler = scheduler
        self.messenger = messenger
        self.lqs_communicator = lqs_communicator
        #self.flask_wrapper = FlaskWrapper(__name__)

        # when debug is enabled => logging messages appear twice
        #self.flask_wrapper.add_endpoint(endpoint='/', endpoint_name="index", handler=self.index)
        #self.flask_wrapper.add_endpoint(endpoint="/index", endpoint_name="index2", handler=self.index)
        #self.flask_wrapper.add_endpoint(endpoint="/trackservice", endpoint_name="trackservice", handler=self.track_service, methods=["GET"])
        #self.flask_wrapper.add_endpoint(endpoint="/api/v1/trackservice/<selected_date>", endpoint_name="api_trackservice", handler=self.api_track_service, methods=["GET"], content_type="application/json", defaults={"selected_date": datetime.date.today()})

        #self.flask_wrapper.run()
        app.run()

@staticmethod
@app.route('/')
@app.route('/index')
@app.route('/index.php')
@app.route('/index.html')
def index():
    return render_template("index.html")

@app.route("/trackservice", methods=["GET"])
def track_service():
    from_time = request.args.get("from")
    to_time = request.args.get("to")
    last = request.args.get("last")
    now = request.args.get("now")

    # nothing set => today's playlist
    if from_time is None and to_time is None and now is None:
        selected_date = datetime.date.today()
        trackservice_entries = TrackService.select_by_day(selected_date)

    # from and end time set
    elif from_time is not None and to_time is not None:
        try:
            to_time = datetime.datetime.strptime(to_time, "%Y-%m-%d")
            from_time = datetime.datetime.strptime(from_time, "%Y-%m-%d")
        except ValueError as e:
            to_time = datetime.datetime.strptime(to_time, "%d.%m.%Y")
            from_time = datetime.datetime.strptime(from_time, "%d.%m.%Y")

        almost_midnight = datetime.time(hour=23, minute=59, second=59)
        to_time = datetime.datetime.combine(to_time, almost_midnight)

        selected_date = str(from_time) + " - " + str(to_time)
        trackservice_entries = TrackService.select_by_range(from_time, to_time)

    # now set
    elif now == "":
        datetime.date.today()
        trackservice_entries = TrackService.now_playing()

#    else
#        return

    return render_template("trackservice.html",
                           length=len(trackservice_entries),
                           trackservice_entries=trackservice_entries,
                           selected_date=selected_date)

@app.route("/api/v1/trackservice", defaults={"from_date": datetime.date.today(), "to_date": None})
@app.route("/api/v1/trackservice/<from_date>")
@app.route("/api/v1/trackservice/<from_date>/")
@app.route("/api/v1/trackservice/<from_date>/<to_date>")
def api_track_service(from_date, to_date=None):

    try:
        # convert date
        try:
            if to_date is not None:
                to_date = datetime.datetime.strptime(to_date, "%Y-%m-%d")
            from_date = datetime.datetime.strptime(from_date, "%Y-%m-%d")
        except ValueError as e:
            if to_date is not None:
                to_date = datetime.datetime.strptime(to_date, "%d.%m.%Y")
            from_date = datetime.datetime.strptime(from_date, "%d.%m.%Y")

        # select from database
        if to_date is None:
            trackservice_entries = TrackService.select_by_day(from_date)
        else:
            trackservice_entries = TrackService.select_by_range(from_date, to_date)

        # return as json
        return simplejson.dumps([tracks._asdict() for tracks in trackservice_entries], default=alchemyencoder)
    except Exception as e:
        import traceback
        traceback.print_exc()

        error = "Cannot transform programme into JSON String. Reason: " + str(e)

        logger = logging.getLogger("AuraEngine")
        logger.error(error)

        return simplejson.dumps({"Error": error})

@staticmethod
@app.route("/api/v1/soundserver_state", methods=["GET"])
def soundserver_settings():
    from modules.communication.liquidsoap.communicator import LiquidSoapCommunicator
    from modules.base.config import ConfigReader

    try:
        cr = ConfigReader()
        cr.load_config()
        lqs = LiquidSoapCommunicator(cr)
        return lqs.auraengine_state()
    except Exception as e:
        error = "Unable to fetch state from Liquidsoap. Is Soundserver running? Reason: " + str(e)
        logger = logging.getLogger("AuraEngine")
        logger.error(error)
        return simplejson.dumps({"Error": error})

@staticmethod
@app.route("/api/v1/upcoming/", methods=["GET"])
def api_clock():
     servertime = datetime.datetime.now()
     # get upcoming tracks
     upcoming = ScheduleEntry.select_upcoming()
     # convert to json string
     upcoming_as_json = simplejson.dumps([tracks._asdict() for tracks in upcoming], default=alchemyencoder)
     # add servertime and return it
     return upcoming_as_json.replace('[{', '[{"servertime":'+str(servertime)+"},{", 1)

@staticmethod
@app.route("/api/v1/sendcommand/<namespace>/")
def send_command(namespace):
    from modules.communication.liquidsoap.communicator import LiquidSoapCommunicator
    from modules.base.config import ConfigReader

    print("Namespace is: "+namespace)

    cr = ConfigReader()
    cr.load_config()
    lqs = LiquidSoapCommunicator(cr)
    x = lqs.__send_lqc_command__(lqs.client, namespace, "");
    print(x)
    return "OK"
